import asyncio
from copy import copy
from inspect import iscoroutinefunction
from typing import List, Union, Any, Coroutine

from loguru import logger

from aflowey import ensure_callable, F, async_wrap


class AsyncFlow:
    """
    Describe an async flow chaining function

    >>>flow = (AsyncFlow() >> gen1 >> gen2 >> gen3)
    >>>await flow.run()
    """

    def __init__(self, *args, **kwargs):
        # first arguments input
        self.args = args
        self.kwargs = kwargs

        # function to be executed in the flow
        self.aws = []

    @staticmethod
    def ensure_coroutine_func(func):
        side_effect_func = hasattr(func, "__side_effect__")
        func = ensure_callable(func)
        if not iscoroutinefunction(func):
            func = F(func) >> async_wrap

        if side_effect_func:
            func.__side_effect__ = True

        return func

    def __rshift__(self, aws) -> "AsyncFlow":
        """add a new step to the flow

        Args:
            aws: list of callable or callable
        """
        if isinstance(aws, list):
            self.aws += [self.ensure_coroutine_func(a) for a in aws]
        else:
            self.aws.append(self.ensure_coroutine_func(aws))

        return self

    def __copy__(self):
        """make a shallow copy"""
        aws = self.aws[:]
        args = self.args[:]
        kwargs = self.kwargs.copy()
        new_flow = self.__class__(*args, **kwargs)
        new_flow.aws = aws
        return new_flow

    async def run(self, **kwargs):
        """run the flow

        Returns:
            coroutine
        """
        return await _FlowExecutor(self).execute_flow(**kwargs)

    @staticmethod
    def log(log_str, print_arg=False):
        """utility function to log between steps, printing argument if needed"""

        async def wrapped(last_result):
            logger.info(log_str)
            if print_arg:  # pragma: no cover
                logger.info(last_result)
            return last_result

        return wrapped

    @staticmethod
    def from_args(*args, **kwargs) -> "AsyncFlow":
        """create a flow with given arguments as first input"""
        return AsyncFlow(*args, **kwargs)

    @staticmethod
    def empty() -> "AsyncFlow":
        """create an empty flow"""
        return AsyncFlow()

    @staticmethod
    def from_flow(flow: "AsyncFlow") -> "AsyncFlow":
        """create a new flow from given flow, copying it
        (args, kwargs, and aws functions)
        """
        assert isinstance(flow, AsyncFlow)
        return copy(flow)


# aliases
aflow = async_flow = AsyncFlow
flog = aflow.log


class _FlowExecutor:
    """Single flow executor"""

    CANCEL_FLOW = object()

    def __init__(self, flow: AsyncFlow):
        self.flow = flow

    @staticmethod
    async def check_and_execute_flow_if_needed(maybe_flow):
        """check if we have an async flow and execute it"""
        if isinstance(maybe_flow, AsyncFlow):
            return await _FlowExecutor(maybe_flow).execute_flow()
        return maybe_flow

    @staticmethod
    def need_to_cancel_flow(result: Any) -> bool:
        if result is _FlowExecutor.CANCEL_FLOW:
            logger.info("Received sentinel object, canceling flow...")
            return True
        return False

    async def execute_flow(self):
        """Main function to execute a flow"""
        if not self.flow.aws:
            logger.debug("no aws")
            return None

        # get first step
        first_aws = self.flow.aws[0]
        if not self.flow.args and hasattr(first_aws, "__F0__"):
            self.flow.args = (None,)

        res = await first_aws(*self.flow.args, *self.flow.kwargs)
        if hasattr(first_aws, "__side_effect__"):
            current_args = self.flow.args[0]
        else:
            current_args = res

        # maybe the step is an async flow
        current_args = await self.check_and_execute_flow_if_needed(current_args)

        # iterate over its tasks
        for task in self.flow.aws[1:]:
            # side effect task, does not return a value
            # reuse the current args
            if hasattr(task, "__side_effect__"):
                result = await task(current_args)
                if self.need_to_cancel_flow(result):
                    break  # pragma: no cover
                continue  # pragma: no cover

            # get the result of a task
            result = await task(current_args)
            result = await self.check_and_execute_flow_if_needed(result)
            # cancel ?
            if self.need_to_cancel_flow(result):
                break
            current_args = result
        # return current args that are the actual results
        return current_args


CANCEL_FLOW = _FlowExecutor.CANCEL_FLOW
FlowOrListFlow = Union[List[AsyncFlow], AsyncFlow]


class AsyncFlowExecutor:
    """
    Execute several flows in parallel

    >>>await (executor(flows) | flow).run()

    """

    def __init__(self, flows: List[AsyncFlow] = [], run_in_thread_pool=False):
        self.flows = []
        self.flows.append(flows)
        # to be run in parallel
        self.run_in_thread_pool = run_in_thread_pool

    def __or__(self, flow: FlowOrListFlow) -> "AsyncFlowExecutor":
        """add a flow to execute in parallel"""
        self.flows.append(flow)

        return self

    @staticmethod
    async def _execute_one_flow(flow: "Flow", *args, **kwargs) -> Any:
        return await _FlowExecutor(flow).execute_flow()

    def exec_or_gather(self, flow: FlowOrListFlow) -> Coroutine:
        execute = self._execute_one_flow
        if isinstance(flow, list):
            return asyncio.gather(
                *[execute(flow, *flow.args, **flow.kwargs) for flow in flow]
            )
        return execute(flow, *flow.args, **flow.kwargs)

    async def run(self, **kwargs):
        """main function to run stuff in parallel"""
        # if other flow run in parallel
        flows = [self.exec_or_gather(flow) for flow in self.flows]
        results = await asyncio.gather(*flows, **kwargs)
        return results

    @staticmethod
    def executor(flows):
        """create a new executor from one flow or array of flows"""
        return AsyncFlowExecutor(flows)


async_exec = aexec = AsyncFlowExecutor.executor



import functools
from inspect import getfullargspec
from inspect import iscoroutine
from typing import Any
from typing import Callable
from typing import List


def identity(x: Any) -> Any:
    return x


class F:
    """tiny wrapper around a function

    Args:
        func: callable

    """

    def __init__(self, func: Callable):
        self.func = func
        functools.update_wrapper(self, func)

    def __rshift__(self, other: Callable) -> "F":
        """equivalent of decoration"""
        return F(other(self.func))

    def __gt__(self, other: Callable) -> "F":
        """function composition"""

        @functools.wraps(other)
        def wrapped(*args, **kwargs):
            return other(self.func(*args, **kwargs))

        return F(wrapped)

    def __call__(self, *args, **kwargs) -> Any:
        """ simply call the inner function"""
        return self.func(*args, **kwargs)

    def __repr__(self) -> str:
        return f"<F instance: {repr(self.func)}>"


def async_wrap(func: Callable) -> F:
    """wrap the given into a coroutine function and try
    calling it
    """

    @functools.wraps(func)
    async def wrapped(*args, **kwargs):
        async def _exec(function, *a, **kw):
            current_result = function(*a, **kw)
            if iscoroutine(current_result):
                current_result = await current_result
                if isinstance(current_result, F):
                    return await _exec(current_result.func)
            return current_result

        return await _exec(func, *args, **kwargs)

    return F(wrapped)


def lift(func, *args, **kwargs) -> F:
    """make a partial function of the given func and ensure
    it will work in an async context
    """
    new_func = functools.partial(func, *args, **kwargs)
    return F(new_func) >> async_wrap


apartial = partial = lift


def f1(func: Callable, extractor: Callable = None) -> F:
    """wraps a one argument function (with arity 1) and allows
        to add an extractor to work on the input argument.
    """

    @functools.wraps(func)
    def wrapped(arg1):
        value = arg1 if extractor is None else extractor(arg1)
        return func(value)
    wrapped.__F1__ = True
    return F(wrapped)


F1 = f1


def f0(func: Callable) -> F:
    """create a new function from a 0 arity function (takes 0 args). The new
     function takes exactly one argument and does not pass it to the wrapped
     function. It allows using a 0 arity function in a flow relatively easily.
    """
    @functools.wraps(func)
    def wrapped(arg1):
        return func()
    wrapped.__F0__ = True
    return F(wrapped)


F0 = f0


def may_fail(func: Callable) -> F:
    """ simply for readability"""
    return func


breaker = erratic = may_fail


def spread_args(func):
    """create a function which takes an iterable of args
    and spread it into the given function"""

    @functools.wraps(func)
    def wrapped(args):
        return func(*args)

    return F(wrapped)


spread = spread_args


def side_effect(*func) -> List:
    """
    take an array of function and tag it as side effects
    function
    """

    def make_impure(f):
        # automatically create new function when the function
        # is a bound method and has no other input args
        f_args = getfullargspec(f).args

        if not isinstance(f, F):
            if not f_args or len(f_args) == 1 and f_args[0] == "self":
                f = F0(f)
            else:
                f = F(f)
        f.__side_effect__ = True
        return f

    if len(func) == 1:
        return make_impure(func[0])
    return [make_impure(fu) for fu in func]


imp = impure = side_effect


def ensure_callable(x: Any) -> Callable:
    if not callable(x):

        def wrapped(*args, **kwargs):
            return x

        return wrapped
    return x
